#include <Adafruit_GFX.h>       //These are all the Libraries needed to make the code function correctly
#include <Adafruit_NeoMatrix.h>
#include <ESP8266WiFi.h>
#include <NTPClient.h>
#include <WiFiUdp.h>
#include <FastLED.h>
#include <math.h>
#include <ArduinoJson.h>
#include <ESP8266HTTPClient.h>

#define PIN D2 // Pin connected to the data input of the LED matrix
#define MATRIX_WIDTH 64
#define MATRIX_HEIGHT 16

#define NUM_LEDS MATRIX_WIDTH * MATRIX_HEIGHT
CRGB leds[NUM_LEDS];

// Replace with your network credentials
const char* ssid = "";
const char* password = "";

// OpenWeather API credentials
const char* apiKey = "";  // Replace with your OpenWeather API key
const char* cityID = "";  // Replace with your city ID

// Define the NTP Client to get the time from Google
WiFiUDP ntpUDP;
const long utcOffsetInSeconds = 3600;  // Adjust this according to your timezone
NTPClient timeClient(ntpUDP, "time.google.com", utcOffsetInSeconds, 60000);

Adafruit_NeoMatrix matrix = Adafruit_NeoMatrix(MATRIX_WIDTH, MATRIX_HEIGHT, PIN,
                                               NEO_MATRIX_TOP + NEO_MATRIX_LEFT + NEO_MATRIX_COLUMNS + NEO_MATRIX_ZIGZAG,
                                               NEO_GRB + NEO_KHZ800);

int brightness = 200;  // Set initial brightness (0 to 255)
int colorIndex = 0;   // Index to keep track of current color

// Array of colors to use for the text and sparkles
uint32_t colors[] = {
  matrix.Color(255, 0, 0),    // Red
  matrix.Color(255, 127, 0),  // Orange
  matrix.Color(255, 255, 0),  // Yellow
  matrix.Color(0, 255, 0),    // Green
  matrix.Color(0, 0, 255),    // Blue
  matrix.Color(75, 0, 130),   // Indigo
  matrix.Color(148, 0, 211)   // Violet
};
const int numColors = sizeof(colors) / sizeof(colors[0]);

// Personal message
String personalMessage = "The Time and Weather on the Rainbow clock is"; // Customize this message
String personalMessage1 = "Have a Nice Day, Whatever the Weather :))"; // Customize this message

// Variables to store weather data
String temperature;
String weatherDescription;

void setup() {
  // Initialize Serial Monitor
  Serial.begin(115200);

  // Connect to Wi-Fi
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(1000);
    Serial.println("Connecting to WiFi...");
  }
  Serial.println("Connected to WiFi");

  // Initialize the NTP client
  timeClient.begin();

  // Initialize the LED matrix
  matrix.begin();
  matrix.setTextWrap(false);
  matrix.setBrightness(brightness);
  
  // Initialize FastLED
  FastLED.addLeds<NEOPIXEL, PIN>(leds, NUM_LEDS);
  FastLED.setBrightness(brightness);
}

// Function to fetch weather data from OpenWeather API
void fetchWeatherData() {
  if (WiFi.status() == WL_CONNECTED) {
    HTTPClient http;
    WiFiClient client;
    String requestURL = "http://api.openweathermap.org/data/2.5/weather?id=" + String(cityID) + "&units=metric&appid=" + String(apiKey);
    http.begin(client, requestURL);
    int httpCode = http.GET();
    
    if (httpCode > 0) {
      String payload = http.getString();
      StaticJsonDocument<1024> jsonBuffer;
      DeserializationError error = deserializeJson(jsonBuffer, payload);
      if (!error) {
        temperature = String(jsonBuffer["main"]["temp"].as<float>(), 1);
        weatherDescription = jsonBuffer["weather"][0]["description"].as<String>();
      } else {
        Serial.println("Failed to parse JSON");
      }
    } else {
      Serial.println("Failed to fetch weather data");
    }
    http.end();
  }
}

// Function to convert (x, y) coordinates to the corresponding LED index
uint16_t XY(uint8_t x, uint8_t y) {
  if (x >= MATRIX_WIDTH || y >= MATRIX_HEIGHT) return -1;
  return (y * MATRIX_WIDTH) + x;
}

class Plasma {
  public:
    Plasma(){};
    bool runPattern();
    void fadeIn();
    void fadeOut();
  private:
    uint16_t _plasmaShift = (random8(0, 5) * 32) + 64;
    uint16_t _plasmaTime = 0;
    const uint8_t _plasmaXfactor = 8;
    const uint8_t _plasmaYfactor = 8;
};

bool Plasma::runPattern() {
  // Fill background with dim plasma
  for (int16_t x = 0; x < MATRIX_WIDTH; x++) {
    for (int16_t y = 0; y < MATRIX_HEIGHT; y++) {
      int16_t r = sin16(_plasmaTime) / 256;
      int16_t h = sin16(x * r * _plasmaXfactor + _plasmaTime) + cos16(y * (-r) * _plasmaYfactor + _plasmaTime) + sin16(y * x * (cos16(-_plasmaTime) / 256) / 2);
      leds[XY(x, y)] = CHSV((uint8_t)((h / 256) + 128), 255, 255);
    }
  }
  uint16_t oldPlasmaTime = _plasmaTime;
  _plasmaTime += _plasmaShift;
  if (oldPlasmaTime > _plasmaTime)
  _plasmaShift = (random8(0, 5) * 32) + 64;

  FastLED.show();
  return true;
}

void Plasma::fadeIn() {
  for (int b = 0; b <= 255; b += 5) {  // Gradually increase brightness
    FastLED.setBrightness(b);
    runPattern();
    delay(30);  // Adjust the delay for the desired fade-in speed
  }
}

void Plasma::fadeOut() {
  for (int b = 255; b >= 0; b -= 5) {  // Gradually decrease brightness
    FastLED.setBrightness(b);
    runPattern();
    delay(30);  // Adjust the delay for the desired fade-out speed
  }
}

void showSparkles(uint16_t duration) {
  uint32_t start = millis();
  while (millis() - start < duration) {
    for (int i = 0; i < 80; i++) {  // 80 random sparkles
      int x = random(MATRIX_WIDTH);
      int y = random(MATRIX_HEIGHT);
      uint32_t color = colors[random(numColors)];
      matrix.drawPixel(x, y, color);
    }
    matrix.show();
    delay(100);  // Slow down the sparkle update frequency
    matrix.fillScreen(0);  // Clear the display
  }
}

void scrollText(String text) {
  int textWidth = text.length() * 6;  // Assuming each character is 6 pixels wide

  // Scroll the text from left to right
  for (int pos = MATRIX_WIDTH; pos >= -textWidth; pos--) {
    matrix.fillScreen(0);                            // Clear the display
    matrix.setCursor(pos, (MATRIX_HEIGHT - 8) / 2);  // Centrally align the text

    // Set text color based on the current color index
    matrix.setTextColor(colors[colorIndex]);
    matrix.print(text);

    // Draw random sparkles
    for (int i = 0; i < 10; i++) {  // 10 random sparkles per frame
      int x = random(MATRIX_WIDTH);
      int y = random(MATRIX_HEIGHT);
      uint32_t color = colors[random(numColors)];
      matrix.drawPixel(x, y, color);
    }

    matrix.show();
    delay(15);  // Adjust this delay to make the text scroll faster or slower/ Lower is faster
  }
}

void loop() {
  // Update the time
  timeClient.update();

  // Get the current time
  String formattedTime = timeClient.getFormattedTime();
  Serial.println(formattedTime);

  // Extract hour from the formatted time
  int currentHour = timeClient.getHours();

  // Check if the current time is within the display hours (7 AM to 11 PM)
  if (currentHour >= 7 && currentHour < 23) {
    // Restore brightness for display hours
    matrix.setBrightness(brightness);

    // Fetch weather data
    fetchWeatherData();

    // Combine the time, weather, and personal messages
    String displayText = personalMessage + " " + formattedTime + " " + "Temp: " + temperature + "C, " + weatherDescription + " " + personalMessage1;

    // Show static rainbow halo using plasma effect with fade-in and fade-out
    Plasma plasma;
    plasma.fadeIn();
    plasma.runPattern();
    plasma.fadeOut();

    // Scroll the combined text with sparkles
    scrollText(displayText);

    // Update the color index to change the text color in the next loop
    colorIndex = (colorIndex + 1) % numColors;
  } else {
    // Dim the display during non-display hours
    matrix.setBrightness(0);
    matrix.show();
  }

  delay(1000);  // Delay before the next update
}
